-- MQTT Client For Control4
-- Written By David Massot 17 Oct 2023
--

local MQTT = {
	TIMEOUT = 10
}

-- Define Client
MQTT.client = {}
MQTT.client.__index = MQTT.client

---
-- Create an MQTT client instance
--
function MQTT.instance(args, callback)

	-- Table: Configuration Arguments
	--        broker: MQTT broker define in Berto Agent
	--        subscriptions: Topics To Subscribe To On Connection
	--        qos: Quality Of Service For Subscritions
	--        timeout: Command Timeout
	--        callback: Function to call when a PUBLISH message is received function(topic, payload)
	-- return:   mqtt_client table

	args = args or {}
			   
	if not type(args) == "table" then
		return
	end
	
	-- Validate Subscriptions
	if args.subscriptions and type(args.subscriptions) == "string" then
		args.subscriptions = { args.subscriptions }
	end
		
	local mqtt_client = {}

	setmetatable(mqtt_client, MQTT.client)

	mqtt_client.broker        = args.broker or "Localhost"
	mqtt_client.timeout       = (args.timeout or MQTT.TIMEOUT) * 1000
	mqtt_client.subscriptions = args.subscriptions
	mqtt_client.qos           = args.qos or 0
	mqtt_client.callback      = callback or function() end
	mqtt_client.seq           = 0
	mqtt_client.queue         = {}
	mqtt_client.connected     = nil
	
	mqtt_client._timer        = {}
	
	return mqtt_client
  
end

-- Handler To Process Data Received From The Broker
function MQTT.client:handler(args)

	if not type(args) == "table" then
		return
	end
	
	if args.SEQ then
		
		local seq = tonumber(args.SEQ)
	
		if not seq then
			return
		end
		
		if self._timer["SEND-" .. seq] then
			self._timer["SEND-" .. seq]:Cancel()
		end
		
		local packet = self.queue[seq]
		
		self.queue[seq] = nil
	
		if not packet then
			return
		end
		
		u.debug("MQTT.client:handler(): Received " .. packet.command .. " With Seq " .. seq, self.broker)
		
		packet.callback(args)
		
	else
		
		local topic   = args.TOPIC
		local message = args.MESSAGE
		
		u.debug("MQTT.client:handler(): Received " .. #message .. " Bytes On " .. topic, self.broker)
		
		self.callback(topic, message)
		
	end
	
end

-- Connect To Broker
function MQTT.client:connect(callback)    
	
	callback = callback or function() end 
	
	self.seq       = 0
	self.queue     = {}
	self.connected = nil
	
	u.debug("MQTT.client:connect()", self.broker)
	
	self:send("MQTT_CONNECT", {
		BROKER = self.broker,
		TOPICS = self.subscriptions and table.concat(self.subscriptions, ","),
		QOS = self.qos,
	}, function(args)
		self.connected = args ~= nil
		callback(args)
	end)


end

function MQTT.client:disconnect(callback)   
	
	callback = callback or function() end 
	
	if not self.connected then
		u.debug("MQTT.client:disconnect(): Not connected", self.broker)
		return
	end
	
	u.debug("MQTT.client:disconnect()", self.broker)
	
	self:send("MQTT_DISCONNECT", nil, function(args)
		self.connected = nil
		callback(args)
	end)

end
  
-- Publish Message
function MQTT.client:publish(topic, message, qos, retain, callback)
	
	callback = callback or function() end 
	
	if not self.connected then
		u.debug("MQTT.client:publish(): Not connected", self.broker)
		return
	end
	
	u.debug("MQTT.client:publish(): " .. topic, self.broker)
	
	self:send("MQTT_PUBLISH", {
		TOPIC = topic,
		MESSAGE = message,
		QOS = qos or 0,
		RETAIN = retain and "True" or "False"
	}, callback)
	
end

function MQTT.client:send(command, params, callback)
	
	params   = params or {}
	callback = callback or function() end
	
	-- Increment Sequence
	if self.seq < 65535 then
		self.seq = self.seq + 1
	else
		self.seq = 1
	end
	
	-- Add Standard Parameters
	params.SEQ    = self.seq
	params.BROKER = self.broker
	params.ID     = C4:GetDeviceID()
	
	-- Add Command To Queue
	self.queue[self.seq] = {
		command = command,
		params = params,
		callback = callback
	}
	
	u.debug("MQTT.client:send(): Sending " .. command .. " With Seq " .. self.seq, self.broker)

	-- Start Timeout Timer
	self._timer["SEND-" .. self.seq] = C4:SetTimer(self.timeout, function()
		u.debug("MQTT.client:send(): Timeout Sending " .. command, self.broker)
		callback()
	end, false)
	
	-- Send Command To Broker
	C4:SendToProxy(999, command, params)
	
end

return MQTT
